package com.tencent.tdmq.rabbitmq.demo.publisherconfirms;

import com.rabbitmq.client.Channel;
import com.rabbitmq.client.ConfirmCallback;
import com.rabbitmq.client.Connection;
import com.rabbitmq.client.ConnectionFactory;
import com.tencent.tdmq.rabbitmq.demo.Constant;

import java.time.Duration;
import java.util.UUID;
import java.util.concurrent.ConcurrentNavigableMap;
import java.util.concurrent.ConcurrentSkipListMap;
import java.util.function.BooleanSupplier;


public class PublisherConfirms {

    static final int MESSAGE_COUNT = 500;

    static Connection createConnection() throws Exception {
        // 连接工厂
        ConnectionFactory factory = new ConnectionFactory();
        // 设置服务地址
        factory.setUri(Constant.URI);
        // 设置Virtual Hosts
        factory.setVirtualHost(Constant.VHOST_NAME);
        // 设置用户名
        factory.setUsername(Constant.USERNAME);
        // 设置密码
        factory.setPassword(Constant.PASSWORD);
        return factory.newConnection();
    }

    public static void main(String[] args) throws Exception {
        // 比较几种发布者确认机制的效率
        publishMessagesIndividually(); //15890
        publishMessagesInBatch(); //346
        handlePublishConfirmsAsynchronously();  // 240
    }

    /**
     * 同步确认
     * @throws Exception
     */
    static void publishMessagesIndividually() throws Exception {
        // 获取连接
        try (Connection connection = createConnection()) {
            // 建立通道
            Channel ch = connection.createChannel();
            // 获取消息队列明白并进行绑定
            String queue = UUID.randomUUID().toString();
            ch.queueDeclare(queue, false, false, true, null);
            // 启用发布者确认
            ch.confirmSelect();
            long start = System.nanoTime();
            for (int i = 0; i < MESSAGE_COUNT; i++) {
                String body = String.valueOf(i);
                // 发布消息
                ch.basicPublish("", queue, null, body.getBytes());
                // 同步等待确认
                ch.waitForConfirmsOrDie(5_000);
            }
            long end = System.nanoTime();
            System.out.format("Published %,d messages individually in %,d ms%n", MESSAGE_COUNT, Duration.ofNanos(end - start).toMillis());
        }
    }

    /**
     * 批量确认消息
     * @throws Exception
     */
    static void publishMessagesInBatch() throws Exception {
        // 获取连接
        try (Connection connection = createConnection()) {
            // 建立通道
            Channel ch = connection.createChannel();
            // 获取消息队列明白并进行绑定
            String queue = UUID.randomUUID().toString();
            ch.queueDeclare(queue, false, false, true, null);
            // 启用发布者确认
            ch.confirmSelect();

            int batchSize = 100;
            int outstandingMessageCount = 0;

            long start = System.nanoTime();
            for (int i = 0; i < MESSAGE_COUNT; i++) {
                String body = String.valueOf(i);
                // 发布消息
                ch.basicPublish("", queue, null, body.getBytes());
                outstandingMessageCount++;

                if (outstandingMessageCount == batchSize) {
                    // 批量确认
                    ch.waitForConfirmsOrDie(5_000);
                    outstandingMessageCount = 0;
                }
            }

            if (outstandingMessageCount > 0) {
                // 批量确认最后一批消息
                ch.waitForConfirmsOrDie(5_000);
            }
            long end = System.nanoTime();
            System.out.format("Published %,d messages in batch in %,d ms%n", MESSAGE_COUNT, Duration.ofNanos(end - start).toMillis());
        }
    }

    /**
     * 异步确认消息
     * @throws Exception
     */
    static void handlePublishConfirmsAsynchronously() throws Exception {
        // 获取连接
        try (Connection connection = createConnection()) {
            // 建立通道
            Channel ch = connection.createChannel();
            // 获取消息队列明白并进行绑定
            String queue = UUID.randomUUID().toString();
            ch.queueDeclare(queue, false, false, true, null);
            // 启用发布者确认
            ch.confirmSelect();
            // 保存已发布待确认的消息信息
            ConcurrentNavigableMap<Long, String> outstandingConfirms = new ConcurrentSkipListMap<>();
            // 确认回调（此处用于清除已发布的消息）
            ConfirmCallback cleanOutstandingConfirms = (sequenceNumber, multiple) -> {
                if (multiple) {
                    ConcurrentNavigableMap<Long, String> confirmed = outstandingConfirms.headMap(
                            sequenceNumber, true
                    );
                    confirmed.clear();
                } else {
                    outstandingConfirms.remove(sequenceNumber);
                }
            };

            // 添加回调  第一个是消息确认的回调   第二分是消息nack的回调
            ch.addConfirmListener(cleanOutstandingConfirms, (sequenceNumber, multiple) -> {
                String body = outstandingConfirms.get(sequenceNumber);
                System.err.format(
                        "Message with body %s has been nack-ed. Sequence number: %d, multiple: %b%n",
                        body, sequenceNumber, multiple
                );
                cleanOutstandingConfirms.handle(sequenceNumber, multiple);
            });

            long start = System.nanoTime();
            for (int i = 0; i < MESSAGE_COUNT; i++) {
                String body = String.valueOf(i);
                // 将发布的消息放到已发送待确认的容器中
                outstandingConfirms.put(ch.getNextPublishSeqNo(), body);
                // 发布消息
                ch.basicPublish("", queue, null, body.getBytes());
            }

            // 等待任务处理结束
            if (!waitUntil(Duration.ofSeconds(60), () -> outstandingConfirms.isEmpty())) {
                throw new IllegalStateException("All messages could not be confirmed in 60 seconds");
            }

            long end = System.nanoTime();
            System.out.format("Published %,d messages and handled confirms asynchronously in %,d ms%n", MESSAGE_COUNT, Duration.ofNanos(end - start).toMillis());
        }
    }

    static boolean waitUntil(Duration timeout, BooleanSupplier condition) throws InterruptedException {
        int waited = 0;
        while (!condition.getAsBoolean() && waited < timeout.toMillis()) {
            Thread.sleep(100L);
            waited = +100;
        }
        return condition.getAsBoolean();
    }

}
