package com.tencent.tdmq.rabbitmq.demo.rpc;

import com.rabbitmq.client.AMQP;
import com.rabbitmq.client.Channel;
import com.rabbitmq.client.Connection;
import com.rabbitmq.client.ConnectionFactory;
import com.tencent.tdmq.rabbitmq.demo.Constant;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.UUID;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;

/**
 * RPC 模型示例
 * RPC客户端
 */
public class RPCClient implements AutoCloseable {

    /**
     * rabbitmq连接
     */
    private final Connection connection;

    /**
     * 连接通道
     */
    private final Channel channel;

    /**
     * RPC请求消息队列名称
     */
    private static final String REQUEST_QUEUE_NAME = "rpc_queue";

    private static final String REPLY_QUEUE = "reply_queue";

    public RPCClient() throws Exception {
        // 连接工厂
        ConnectionFactory factory = new ConnectionFactory();
        // 设置服务地址
        factory.setUri(Constant.URI);
        // 设置Virtual Hosts
        factory.setVirtualHost(Constant.VHOST_NAME);
        // 设置用户名
        factory.setUsername(Constant.USERNAME);
        // 设置密码
        factory.setPassword(Constant.PASSWORD);
        // 创建连接
        connection = factory.newConnection();
        // 创建消息通道
        channel = connection.createChannel();
    }

    public static void main(String[] argv) {
        try (RPCClient fibonacciRpc = new RPCClient()) {
            for (int i = 0; i < 32; i++) {
                String requestParam = Integer.toString(i);
                System.out.println(" [Send RPC request] Requesting fib(" + requestParam + ")");
                String response = fibonacciRpc.call(requestParam);
                System.out.println(" [Get RPC response] Got '" + response + "'");
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public String call(String message) throws IOException, InterruptedException {
        final String corrId = UUID.randomUUID().toString();
        // 传递参数绑定回调信息队列，设置请求id
        AMQP.BasicProperties props = new AMQP.BasicProperties
                .Builder()
                .correlationId(corrId)
                .replyTo(REPLY_QUEUE)
                .build();
        // 发起RPC调用
        channel.basicPublish("", REQUEST_QUEUE_NAME, props, message.getBytes(StandardCharsets.UTF_8));

        final BlockingQueue<String> response = new ArrayBlockingQueue<>(1);
        // 设置回调队列消费者，用来获取回调信息
        String ctag = channel.basicConsume(REPLY_QUEUE, true, (consumerTag, delivery) -> {
            if (delivery.getProperties().getCorrelationId().equals(corrId)) {
                response.offer(new String(delivery.getBody(), StandardCharsets.UTF_8));
            }
        }, consumerTag -> {
        });
        // RPC请求结果
        String result = response.take();
        // RPC结束，关闭消费者
        channel.basicCancel(ctag);
        return result;
    }

    @Override
    public void close() throws IOException {
        connection.close();
    }
}
